const { Log } = require('./Log')
const { Wire } = require('./Wire')
const { SerialPort } = require('serialport')

/**
 * # Board
 * 
 * This is the main class that will take care of the communication with the Arduino board.
 * 
 * ```
 * const { Board } = require('./modules/nnarduino');
 * 
 * // auto-detect board?
 * let board = new Board();
 * 
 * // or set board manually?
 * let board = new Board({path:'/dev/tty.usbserial-A9007KK5', baudRate:57600});
 * 
 * // connect to the board, and wait for first real "handshake". The Board is then ready to communicate.
 * board.connect().then(...).catch(...);
 * ```
 * 
 */
class Board {
	
	STATE_DISCONNECTED = -1;
	STATE_CONNECTING = 0;
	STATE_CONNECTED = 1;

	port = null;
	wire = null;
	features = {};
	state = -1;

	path = '';
	baudRate = 9600;
	eventListeners = {};
	
	/**
	 * Returns `TRUE` if the board is connected and handshake was successful
	 * 
	 * @returns {boolean}
	 */
	get connected() {
		return this.state == this.STATE_CONNECTED;
	}

	/**
	 * These patterns for the port.path will be checked during autoDetect in order to decide, 
	 * if the connected device is an Arduino board. Needs to be improved.
	 * 
	 * @type {array}
	 */
	patternsToCheck = [
		'(.*)tty\.usbserial(.*)',
		'(.*)tty\.usbmodem(.*)',
		'(.*)ttyACM(.*)',
		'(.*)ttyUSB(.*)',
		'(.*)ttyS(.*)',
		'(.*)COM(0-9*)(.*)',
		'COM([0-9]*)(.*)',
		'COM([0-9]*)',
	];

	/**
	 * ... and these port.manufacturers will be checked.
	 * 
	 * @type {array}
	 */
	manufacturersToCheck = [
		'FTDI',
		'arduino'
	];

	/**
	 * Constructor.
	 * 
	 * @param {object} config
	 * @returns {void}
	 */
	constructor( config = {} ) 
	{
		if (config.path) {
			this.path = config.path;
		}
		if (config.baudRate) {
			this.baudRate = config.baudRate;
		}
		this.wire = new Wire({
			board: this
		});
	}

	/**
	 * Adds an eventListener to the Board
	 * 
	 * @returns void
	 */
	on( event, callback ) 
	{
		if (!this.eventListeners[event]) {
			this.eventListeners[event] = [];
		}
		this.eventListeners[event].push( callback );
	}
	
	/**
	 * Removes an eventListener
	 * 
	 * @returns void
	 */
	off( event, callback ) 
	{
		if (!this.eventListeners[event]) return;
		this.eventListeners[event] = this.eventListeners[event].filter( func => func != callback );
		console.log(this.eventListeners);
	}

	/**
	 * Emit an event
	 * 
	 * @returns void
	 */
	emit( event, params ) 
	{
		this.eventListeners[event]?.forEach(( func ) => {
			func( params );
		});
	}

	/**
	 * Connect to device and create listeners
	 * 
	 * @return {Promise}
	 */
	async connect( path = null ) 
	{
		if (path != this.path) {
			this.disconnect()
			this.state = this.STATE_DISCONNECTED
			this.path = path
		}

		if (this.state != this.STATE_DISCONNECTED) {
			return true;
		}

		this.state = this.STATE_CONNECTING;
		let autoDetectedPaths = []

		try {
			autoDetectedPaths = await this.autoDetect();
		} catch( paths ) {
			this.emit('error', {message:'No board detected.'})
			Log.ERROR( 
				'Board->autoDetect()', 
				//'No board detected.')
				'No board detected. None of the following paths: \n\n• ' + paths.join('\n• ') + '\n\n matched any of the RegEx defined in class `Board` \n\n• ' + this.patternsToCheck.join('\n• '));
			throw new Error('No board detected');
		}

		if (this.path) {
			autoDetectedPaths = autoDetectedPaths.filter( p => p != this.path )
			autoDetectedPaths.push( this.path )
		}

		while (autoDetectedPaths.length > 0) {
			this.path = autoDetectedPaths.pop()
			try {
				await this.openPort()
				break;
			} catch ( error ) {
				this.emit('error', {message:`Open port ${this.path} failed.`})
				this.path = false
			}	
		}

		if (!this.path) {
			this.emit('error', {message:`Could not connect to device.`})
			throw new Error('Open port failed.');
		}

		return true;
	}

	/**
	 * Open the port and create the virtual Wire
	 * 
	 * @returns {Promise}
	 */
	openPort() 
	{
		return new Promise((resolve, reject) => {

			this.port = new SerialPort({
				path: this.path, 
				baudRate: this.baudRate,
				autoOpen: true,
				lock: false,
			});

			this.port.on('open', () => {

				// make sure to disconnect, if window is reloaded / refreshed
				window.onbeforeunload = this.disconnect.bind(this);

				Log.SUCCESS('Board->connect()', `port to ${this.path} open`);
				this.wire.connect( this.port );
				//this.emit('portOpened', {port:this.port, path:this.path})
				this.emit('status', {text:`Trying to connect to ${this.path} ...`})

				// handshake
				this.wire.handshake().then(( data ) => {
					this.features = data;
					Log.SUCCESS('Board->handshake()', data);
					this.emit('handshake', {data, port:this.port, path:this.path})
					this.state = this.STATE_CONNECTED;
					this.wire.processQueue();
					resolve();
				}).catch( reject );
			});

			this.port.on('error', (err) => {
				Log.ERROR( 'Board->connect()', err.message )
				this.emit('error', err)
				if (err.message.indexOf('Resource busy') > -1) {
					alert('Unable to connect to Arduino:\nSerial bus is busy.\n\nPlease close the Arduino serial monitor, if it is open!');
				}
				reject( err.message );
			})

			this.port.on('close', (err) => {
				Log.SUCCESS( 'Board->connect()', 'closed connection' )
				this.port = null
				this.state = Wire.STATE_DISCONNECTED
				this.emit('portClosed')
			})
		})
	}

	/**
	 * Disconnect from board 
	 * 
	 * @returns {void}
	 */
	disconnect() 
	{
		if (!this.port) return;
		this.wire.disconnect()
		Log.INFO( 'Board->disconnect()', 'disconnecting...' )
		this.port.close()
	}

	async listDevices() 
	{
		var ports = await SerialPort.list()
		return ports
	}

	/**
	 * Autodetect Aduino devices by pattern matching paths.
	 * 
	 * @returns {Promise}
	 */
	async autoDetect() 
	{
		var ports = await SerialPort.list();
		let paths = [];
		let arduinos = [];

		ports.forEach((port) => {
			paths.push( port.path );					
			this.patternsToCheck.forEach((pattern) => {
				if (port.path.match( new RegExp(pattern, 'i') )) {
					arduinos.push( port.path );
				}
			});
			this.manufacturersToCheck.forEach((pattern) => {
				if (port.manufacturer?.match( new RegExp(pattern, 'i') )) {
					arduinos.push( port.path );
				}
			});
		});

		if (!arduinos.length) throw paths;
		return arduinos;
	}

	/**
	 * Send a command to Arduino accross the "Wire".
	 * 
	 * Examples:
	 * ```
	 * board.cmd({blink:true});
	 * board.cmd({setIV:10000});
	 * board.cmd({addSensor:{pin:14, type:20}});
	 * board.cmd({addSensor:{pin:"A0", type:20}});
	 * board.cmd({pinMode:{pin:13, mode:1}});
	 * board.cmd([{digitalWrite:{pin:13, mode:1}}], {delay:1000});
	 * ```
	 * @return {Promise}
	 */
	cmd( obj = {}, options = {} ) 
	{
		return this.wire.createRequest( obj, options );
	}

}

exports.Board = Board;