const { Uniq } = require('./Uniq')

/**
 * # Request
 * 
 * Request-Object to be sent over the Wire to Arduino
 * 
 * ```
 * const { Request } = require('./modules/nnarduino');
 * 
 * ```
 * 
 */
class Request {

	id = null;
	message = null;
	options = {};

	/**
	 * Constructor.
	 * 
	 * @param {object} obj
	 * @param {object} options
	 * @returns {self}
	 */
	constructor( obj = {}, options = {} ) 
	{
		options = Object.assign({
			iv: 3000,
			attempts: 3,
			delay: 0,
			resolve: false,
			reject: false,
		}, options);

		this.id = Uniq.id;
		this.message = this.createCommand( this.id, obj );
		this.options = options;
	}

	/**
	 * 
	 * ```
	 * this.command({blink:true});
	 * // --> {id:123, cmd:[{blink:true}]}
	 * 
	 * this.command([{blink:true}, {blink:true}]);
	 * // --> {id:123, cmd:[{blink:true}, {blink:true}]}
	 * ```
	 * @param {int} id
	 * @param {mixed} cmd
	 * @returns {object}
	 */	
	createCommand( id = 0, cmd = {} ) 
	{
		 if (!Array.isArray(cmd)) cmd = [cmd];
		 return {id:id, cmd:cmd}
	}

	/**
	 * Convert the message to a Buffer so it can be sent via SerialPort
	 * ```
	 * let buffer = request.toBuffer();
	 * // --> <Buffer> '{"a":1, "b":2}\n'
	 * ```
	 * 
	 * @param {object} message 
	 * @returns {Buffer}
	 */
	toBuffer() 
	{
		let json = this.toString();
		return Buffer.from( json + "\n" );
	}
	
	toString() 
	{
		return JSON.stringify( this.message )
	}

	applyBoardFeatures( features = {} ) 
	{
		this.parsePinNumbers( this.message.cmd, features );
	}

	/**
	 * Convert `A0` to numeric pinNumber
	 * 
	 * @return {array}
	 */
	parsePinNumbers( objArr = {}, features = {} ) 
	{
		const pinKeysToParse = ['pin', 'vry', 'vrx'];

		objArr.forEach(( obj ) => {
			for (let cmd in obj) {
				let params = obj[cmd];
				if (typeof params !== 'object') continue;
				for (let k in params) {
					let v = params[k];
					if (pinKeysToParse.indexOf(k) == -1) continue;
					if (`${v}`.charAt(0) == 'A') {
						params[k] = parseInt( v.substr(1) ) + parseInt( features.A0 );
					}
				}
			}
		});

		return objArr;
	}

}

exports.Request = Request;