const { Log } = require('./Log')
const { Request } = require('./Request')
const { ArduinoEvent } = require('./ArduinoEvent')
const { ReadlineParser } = require('@serialport/parser-readline');

const DEBUG = false;

/**
 * # Wire
 * 
 * Takes care of sending, receiving and parsing messages from and to Arduino
 * 
 * ```
 * const { Message } = require('./modules/nnarduino');
 * 
 * ```
 * 
 */
class Wire {
	
	port = null;
	parser = null;
	board = null;

	checkQueueIV = false;

	/**
	 * The requests waiting to be sent to Arduino.
	 * These will be sent after each other - one by one.
	 * 
	 * @type {array}
	 */
	requestQueue = [];

	/**
	 * The requests that have been sent to Arduino and are
	 * currently waiting for a response.
	 * 
	 * @type {object}
	 */
	pendingRequests = {};

	/**
	 * Constructor.
	 * 
	 * @param {object} config
	 * @returns {void}
	 */
	constructor( config = {} ) 
	{
		if (config?.port) {
			this.connect( config.port );
		}
		this.board = config.board;
	}

	/**
	 * 
	 * @param {object} port
	 * @returns {self}
	 */
	connect( port = false ) 
	{
		this.port = port;
		let parserInstance = new ReadlineParser({ 
			//delimiter: Buffer.from(String.fromCharCode(END_MESSAGE), 'latin1'),
			//encoding: 'latin1' 
		});
		this.parser = this.port.pipe( parserInstance );
		this.parser.on('data', this.parseResponse.bind( this ) );
		//this.port.on('data', this.parseResponse.bind( this ) )
		return this;
	}

	/**
	 * Disconnect from board 
	 * 
	 * @returns {void}
	 */
	 async disconnect() 
	 {
		 if (!this.parser) return;
		 Log.INFO( 'Wire->disconnect()', 'disconnecting...' )
		 this.parser.off('data', this.parseResponse.bind( this ) )
		 this.port = null
	 }

	/**
	 * Check if there are requests in the queue waiting to be sent.
	 * Send next message waiting.
	 * 
	 * @returns {boolean}
	 */
	processQueue( forceNext = false ) 
	{
		// handshake not completed, abort
		if (!this.board.connected) return false;

		// nothing to process?
		if (!this.requestQueue.length) return false;

		// is current request still pending?
		let firstMessageId = this.requestQueue[0].id;
		if (this.pendingRequests[firstMessageId]) return false;

		let request = this.requestQueue.shift();

		this.send( request ).then( request.options.resolve?.() ).catch( request.options.reject?.() );

		return true;
	}

	/**
	 * Create a command-request to the queue.
	 * 
	 * @param {object} obj
	 * @returns {Promise}
	 */
	 createRequest( obj = {}, options = {} ) 
	 {
		if (!Array.isArray(obj)) {
			obj = [obj];
		}

		return new Promise((resolve, reject) => {
			options = Object.assign({
				resolve,
				reject,
			}, options);
			let request = new Request( obj, options );
			this.requestQueue.push( request );
			this.processQueue();
		});
	}

	/**
	 * 
	 * @returns {object}
	 */

	/**
	 * Handshake with board, wait for SerialPort on Arduino to respond to messages sent.
	 * 
	 * @returns {Promise}
	 */
	handshake() 
	{
		let request = new Request({'handshake':true}, {delay:1600, iv:500, attempts:5});
		return this.send( request );
	}

	/**
	 * Send request
	 * 
	 * @param {object} item
	 * @returns {Promise}
	 */
	send( request ) 
	{
		return new Promise((resolve, reject) => {

			request.applyBoardFeatures( this.board.features );

			let id = request.id;
			let me = this;

			this.pendingRequests[id] = {
				request,
				resolve,
				reject
			}
			
			let attempts = request.options.attempts;
			function send() {
				if (!me.pendingRequests?.[id]) return;
				if (attempts-- > 0) {
					me.port.write( request.toBuffer(), function( err ) {
						if (err) {
							delete(me.pendingRequests[id]);
							Log.ERROR('Error on write: ', err.message);
							return reject();
						}
						if (DEBUG) Log.INFO( `Request sent [${id}]`, request.toString() );
					})	
					setTimeout( send, request.options.iv )
				} else {
					if (me.pendingRequests[id]?.options?.reject) {
						me.pendingRequests[id].options.reject(`aborted`);
					}
					delete(me.pendingRequests[id]);
					me.board.emit('error', {message:'Aborted sending message.'})
					Log.ERROR(`Aborted sending message after ${request.options.attempts} attempts.`);
					reject('Handshake failed.');
				}
			}
			if (!request.options.delay) {
				send();
			} else {
				setTimeout( send, request.options.delay );
			}
		});
	}

	/**
	 * Handle response
	 * 
	 * @param {Buffer} data
	 * @return void
	 */
	parseResponse( json ) 
	{	
		let data = {};

		try {
			data = JSON.parse( json );
		} catch( e ) {
			return Log.ERROR('Unparsable response from Arduino', json );
		}

		// add pinNumber-alias (14 -> `A0`) to result-Array `{14:{...}, ...}` => {14:{...}, A0:{...}, ...}
		if (data.pins) {
			for (let pinNum in data.pins) {
				let ref = data.pins[pinNum];
				let pinIdentifer = this.pinNumToIdentifer(pinNum);
				data.pins[pinIdentifer] = ref;
			}
		}

		if (data.error) {
			return Log.ERROR('Error from Arduino', data.message );
		}
		
		if (data.id) {
			let request = this.pendingRequests[data.id];
			request?.resolve?.( data );
			if (DEBUG) Log.SUCCESS(`Arduino processed request [${data.id}]`, data );
			delete( this.pendingRequests[data.id] );
			this.processQueue();
		} else if (data.event) {
			let event = new ArduinoEvent( data );
			this.board.emit( event.type, event );
			if (DEBUG) Log.INFO(`Event '${event.type}' from Arduino`, event );
		} else if (data) {
			if (DEBUG) Log.INFO('Response from Arduino', data );
		}
	}

	/**
	 * Converts a pin-number (i.e. `14`) to the corresponding identifer (`A0`)
	 * 
	 * @param {int|string} pinNum
	 * @returns {string}
	 */
	pinNumToIdentifer( pinNum ) 
	{
		pinNum *= 1;
		if (pinNum < this.board.features.A0) {
			return pinNum;
		}
		return `A` + (pinNum - this.board.features.A0);
	}
	 
}

exports.Wire = Wire;